# SPDX-FileCopyrightText: Copyright (c) 2021 Dan Halbert for Adafruit Industries
#
# SPDX-License-Identifier: MIT

# pylint: disable=line-too-long
"""
`adafruit_radial_controller.device`
================================================================================

* Author(s): Dan Halbert

The radial controller report descriptor used is described by Microsoft
`here <https://docs.microsoft.com/en-us/windows-hardware/design/component-guidelines/radial-controller-sample-report-descriptors>`_.
"""
# pylint: enable=line-too-long

__version__ = "1.0.4"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_Radial_Controller.git"

import usb_hid


def device(report_id: int) -> usb_hid.Device:
    """Create a `usb_hid.Device` for a radial controller.

    :param int report_id: The report ID to use for the device.
    """
    return usb_hid.Device(
        # fmt: off
        report_descriptor=bytes((
            0x05, 0x01,       # Usage Page (Generic Desktop)
            0x09, 0x0e,       # Usage (System Multi-Axis Controller)
            0xa1, 0x01,       # Collection (Application)
            0x85, report_id,  #   Report Id (Radial Controller)
            0x05, 0x0d,       #   Usage Page (Digitizers)
            0x09, 0x21,       #   Usage (Puck)
            0xa1, 0x00,       #   Collection (Physical)

            # The Microsoft example packs the report into 1 bit and 15 bits,
            # but that's more work. We'll just use 1 bit of one byte and then 16 bits.

            # Button
            0x05, 0x09,       #     Usage Page (Buttons)
            0x09, 0x01,       #     Usage (Button 1)
            0x15, 0x00,       #     Logical Minimum (0)
            0x25, 0x01,       #     Logical Maximum (1)
            0x75, 0x01,       #     Report Size (1)
            0x95, 0x01,       #     Report Count (1)
            0x81, 0x02,       #     Input (Data,Var,Abs)

            # Padding
            0x75, 0x07,       #     Report Size (7)
            0x95, 0x01,       #     Report Count (1)
            0x81, 0x01,       #     Input (Data,Var,Abs)

            # Rotation
            0x05, 0x01,       #     Usage Page (Generic Desktop)
            0x09, 0x37,       #     Usage (Dial)
            0x55, 0x0f,       #     Unit Exponent (-1)
            0x65, 0x14,       #     Unit (Degrees, English Rotation)
            0x36, 0xf0, 0xf1, #     Physical Minimum (-3600)
            0x46, 0x10, 0x0e, #     Physical Maximum (3600)
            0x16, 0xf0, 0xf1, #     Logical Minimum (-3600)
            0x26, 0x10, 0x0e, #     Logical Maximum (3600)
            0x75, 0x10,       #     Report Size (16)
            0x95, 0x01,       #     Report Count (1)
            0x81, 0x06,       #     Input (Data,Var,Rel)

            0xc0,             #   End Collection
            0xc0,             # End Collection

            # The presence of this (empty) System Control Collection makes the device
            # appear in /dev/input in Linux, which will help if radial controller
            # functionality is added in userspace code.
            0x09, 0x80,       # Usage (System Control)
            0xa1, 0x01,       # Collection (Application)
            0xc0,             # End Collection
            )),
        # fmt: on
        usage_page=0x01,
        usage=0x0E,
        report_ids=(report_id,),
        in_report_lengths=(3,),
        out_report_lengths=(0,),
    )



# SPDX-FileCopyrightText: Copyright (c) 2021 Dan Halbert for Adafruit Industries
#
# SPDX-License-Identifier: MIT
"""
`adafruit_radial_controller`
================================================================================

HID Radial Controller device helper library


* Author(s): Dan Halbert

Implementation Notes
--------------------

**Hardware:**

  At the minimum, a radial controller is a rotary encoder plus a switch.

  Documentation is available from Microsoft:
  https://docs.microsoft.com/en-us/windows-hardware/design/component-guidelines/radial-implementation-guide

**Software and Dependencies:**

* Adafruit CircuitPython firmware for the supported boards:
  https://github.com/adafruit/circuitpython/releases
"""
import time

import usb_hid
from adafruit_hid import find_device

__version__ = "1.0.4"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_Radial_Controller.git"


class RadialController:
    """Send Radial Controller HID reports."""

    def __init__(self, devices):
        """Create a RadialController object that will send Radial Controller HID reports.

        Devices can be a list of devices that includes a keyboard device or a keyboard device
        itself. A device is any object that implements ``send_report()``, ``usage_page`` and
        ``usage``.
        """
        self._controller = find_device(devices, usage_page=0x01, usage=0x0E)

        # Reuse this bytearray to send radial controller reports
        # report[0]: bit 0: button
        # report[1]: rotation
        self.report = bytearray(3)

        self._pressed = False

        # Do a no-op to test if HID device is ready.
        # If not, wait a bit and try once more.
        try:
            self._send(False, 0)
        except OSError:
            time.sleep(1)
            self._send(False, 0)

    def press(self):
        """Press the button."""
        self._pressed = True
        self._send(True, 0)

    def release(self):
        """Release the button."""
        self._pressed = False
        self._send(False, 0)

    def click(self):
        """Press and release the button."""
        self.press()
        self.release()

    def rotate(self, degree_tenths):
        """Set relative rotation value, in tenths of a degree.
        A value of +/- 1 or 10 can be too small and cause tool selection or scrolling to not work.
        +/- 100 is a good value for a single increment in many cases, though it causes value sliders
        to change by 10 instead of 1.
        """

        if not -3600 <= degree_tenths <= 3600:
            raise ValueError("rotation must be in range -3600 to 3600")
        self._send(self._pressed, degree_tenths)

    def _send(self, pressed, rotation):
        self.report[0] = int(pressed)
        self.report[1] = rotation & 0xFF
        self.report[2] = (rotation >> 8) & 0xFF

        self._controller.send_report(self.report)



# SPDX-FileCopyrightText: Copyright (c) 2021 Dan Halbert for Adafruit Industries
#
# SPDX-License-Identifier: Unlicense

import board
import digitalio
import rotaryio
import usb_hid

from adafruit_debouncer import Debouncer
import adafruit_radial_controller

switch = digitalio.DigitalInOut(board.SWITCH)
switch.pull = digitalio.Pull.DOWN
debounced_switch = Debouncer(switch)

encoder = rotaryio.IncrementalEncoder(board.ROTA, board.ROTB)

radial_controller = adafruit_radial_controller.RadialController(usb_hid.devices)

last_position = 0
DEGREE_TENTHS_MULTIPLIER = 100

while True:
    debounced_switch.update()
    if debounced_switch.rose:
        radial_controller.press()
    if debounced_switch.fell:
        radial_controller.release()

    position = encoder.position
    delta = position - last_position
    if delta != 0:
        radial_controller.rotate(delta * DEGREE_TENTHS_MULTIPLIER)
        last_position = position



# SPDX-FileCopyrightText: Copyright (c) 2021 Dan Halbert for Adafruit Industries
#
# SPDX-License-Identifier: Unlicense

# You must use a board-specific example to do a simple test.
#
# radial_controller_rotary_trinkey.py is an example.



# SPDX-FileCopyrightText: Copyright (c) 2021 Dan Halbert for Adafruit Industries
#
# SPDX-License-Identifier: Unlicense

import usb_hid

import adafruit_radial_controller.device

REPORT_ID = 5

radial_controller_device = adafruit_radial_controller.device.device(REPORT_ID)
usb_hid.enable((radial_controller_device,))
